#!/usr/bin/perl
#
# $Header: //sapdb/V73/c_00/b_43/sys/src/install/perl/SAPDB/Install/Misc.pm#1 $
# $DateTime: 2003/08/28 15:24:28 $
# $Change: 50893 $
#
# Desc: 


package SAPDB::Install::Misc;

$VERSION = 1.01;

sub BEGIN {
      @ISA = ('SAPDB::Install::Exporter');
      @EXPORT = ('printVersion','selectInstallation','newinstall','update','delete','patch','verify','getPath','checkOwner','checkGroup','checkNode','expandPath');

	my $repo = SAPDB::Install::Repository::GetCurrent ();
	my @neededPackages=(
		'StdIO',
		'Values',
		'Getopt::Long',
		'Tools',
		'System',
		'Trace'
	);
	foreach my $package (@neededPackages) {
	  	unless (defined $repo->Eval ("SAPDB::Install::$package", 1.01)) {
                print join ("\n", $repo->GetErr)."\n";
                die;
        	}
		SAPDB::Install::Exporter::import ("SAPDB::Install::$package");
	  } 
}





sub printVersion{
	local (*ARGV)=@_;
	%opt_ctrl=(
		'v' => \$opt_v,
		'version' => \$opt_version

	);
	SAPDB::Install::Getopt::Long::Configure('pass_through');
	SAPDB::Install::Getopt::Long::Configure('noauto_abbrev');
	GetOptions(%opt_ctrl);
	if(defined $opt_version || defined $opt_v){
		print "\n\n\nPERL MODULE ARCHIVE VERSION: $SAPDB::Install::Version::Version (changelist id: $SAPDB::Install::Version::Change)\n";
		print "\n\n\nSDBRUN VERSION: ".$SAPDB::Install::Config{'BinaryInstVersion'}." (changelist id: ".$SAPDB::Install::Config{'BinaryChangeList'}.")\n";
		die ("\n\n\n");
	}
	1;
}





#sub get installation path in interaction with user
sub getDir{
    my $default=$_[0];
	my $path=readstdin();
    $path=~s/\\/\//g;
	unless($path){
		$path=$default;
	}
    if(-d $path){	
		return $path;
   }
    else{
		print2stdout("directory \"$path\" does not exist, create? (y/n) ");
		$_=readstdin();
		if(/^[yY].*/){
			return $path;
		}
		else{
			print2stdout("abort installation? (y/n) ");
			$_=readstdin();
			if(/^[yY].*/){
				diesoft("installation of SAP DB software aborted "); 
			}
		}	
	}
	return "";
}


	
#sub expand relativ path names to absolute
sub expandPath{
	my ($tmp)=@_;
	defined $tmp or return;
	$tmp=~/^\s*$/ and return;
	$tmp=~s/^\s*//g;
	if($^O =~ /mswin/i){
		unless($tmp =~ /^[a-zA-Z]:[\/\\].*/){
			$tmp=$SAPDB::Install::Values::curDir."/$tmp";
		}
	}
	else{
		unless($tmp =~ /^[\/\\].*/){
			$tmp=$SAPDB::Install::Values::curDir."/$tmp";
		}
	}
	return $tmp;
}


sub getPath{
	my ($name,$default,$opt,$batch)=@_;
	my $returnvalue;
	if($opt){
		$returnvalue=expandPath($opt);
		$returnvalue =~ s/\\/\//g;
		return $returnvalue; 
	}
	$batch and print2stderr("run in batch mode - \"$name\" unknown\n") and diesoft($SAPDB::Install::Values::diemsg);
	while(! $returnvalue){
		print2stdout("please enter $name [$default]: ");
		$returnvalue = getDir($default);
	}
	return expandPath($returnvalue);		
}
	

sub checkOwner{
	($^O=~/MSWin.*/i) and return 0;
	my ($default,$packobj,$opt,$batch)=@_;
	sub setPassword{
		 my ($owner) = @_;
		 if($batch){
			print2stderr("WRN: run in batchmode - please set password for os user $owner with command \"passwd $owner\" after installation\n");
		 }
		 else{
			my $outtext=`passwd $owner`;
			# match: linux, hp, sun - success |  true64 - updated | aix - ^$
			$outtext=~/success|updated|^$/gi and $SAPDB::Install::Values::log->SetMsg("MSG: change os user password successfully\n") or print2stdout("cannot set os user password\n");
		}
	}
	if($opt){
		#owner as command line argument
		my $owner=$opt;
		my $uid= (getpwnam($owner))[2];
		if($uid eq ''){
			#user does not exist - try to create new one
			if($^O=~/aix/i){
				callsystem('mkuser pgrp='.$packobj->Group." $owner");
			}
			else{
				callsystem("useradd -c \"owner of SAPDB ".$packobj->DispName."\" -g ".$packobj->Group." -m $owner");# 2> /dev/null > /den/null";
			}
			$uid = (getpwnam($owner))[2];
			if($uid eq ''){	
				#cant create
				print2stderr("cannot create  user \"$owner\"\n");
				diesoft($SAPDB::Install::Values::diemsg);
			}
			setPassword($owner);
		}
		$packobj->User($owner);
		$packobj->UID($uid);
	}
	else{
		$batch and print2stderr("run in batch mode - \"owner\" unknown\n") and diesoft($SAPDB::Install::Values::diemsg);
		while(1){
			print2stdout("please enter owner name for database programs [$default]: ");
			$owner=readstdin();
			unless($owner){
				$owner=$default;
			}
			$uid = (getpwnam($owner))[2];
			if($uid ne ''){
				$packobj->User($owner);
				$packobj->UID($uid);
				last;
			}
			print2stdout("unknown user -  create \"$owner\" on local machine? (y/n) ");
			$_=readstdin();
			if(/^[yY].*/){
				#create new user
				if($^O=~/aix/i){
					callsystem('mkuser pgrp='.$packobj->Group." $owner");
				}
				else{
					callsystem("useradd -c \"owner of SAPDB ".$packobj->DispName."\" -g ".$packobj->Group." -m  $owner");#2> /dev/null > /den/null";
				}
				$uid=(getpwnam($owner))[2];
				if($uid ne ''){
					setPassword($owner);
					$packobj->User($owner);
					$packobj->UID($uid);
					last;
				}
				else{
					print2stderr("cannot create user \"$owner\" \n");
					diesoft($SAPDB::Install::Values::diemsg);
				}
			}
			print2stdout("abort installation? (y/n) ");
			$_=readstdin();
			/^[yY].*/ and diesoft("installation aborted ");
		}
	}
}
	


sub checkGroup{
	($^O=~/MSWin.*/i) and return 0;
	my ($default,$packobj,$opt,$batch)=@_;
	if($opt){
		#group as command line argument
		my $group=$opt;
		my $gid=(getgrnam($group))[2];
		if($gid eq ''){
			#group does not exist - try to create new one
			if($^O=~/aix/i){
				callsystem("mkgroup $group");
			}
			else{
				callsystem("groupadd $group");
			}
			$gid=(getgrnam($group))[2];
			if($gid eq ''){
				#cant create group
				print2stderr("cannot create group\n");
				diesoft($SAPDB::Install::Values::diemsg);
			}
		}
		$packobj->Group($group);
		$packobj->GID($gid);
	}
	else{
		$batch and print2stderr("run in batch mode - \"group\" unknown\n") and diesoft($SAPDB::Install::Values::diemsg);
		while(1){
			print2stdout("please enter group name for database programs [$default]: ");
			my $group=readstdin();
			unless($group){
				$group=$default;
			}
			my $gid=(getgrnam($group))[2];
			if($gid ne ''){
				$packobj->Group($group);
				$packobj->GID($gid);
				last;
			}
			print2stdout("unknown group - create \"$group\" on local machine? (y/n) ");
			$_=readstdin();
			if(/^[yY].*/){
				#create new group
				if($^O=~/aix/i){
					callsystem("mkgroup $group");
				}
				else{
					callsystem("groupadd $group");
				}
				$gid=(getgrnam($group))[2];
				if($gid ne ''){
					$packobj->Group($group);
					$packobj->GID($gid);
					last;
				}
				else{
					print2stderr("cannot create group \"$group\" \n");
					diesoft($SAPDB::Install::Values::diemsg);
				}
			}
	
			print2stdout("abort installation? (y/n) ");
			$_=readstdin();
			/^[yY].*/ and diesoft("installation aborted ");
		}
	}
}


sub newinstall {
	my ($packobj)=@_;
	my $regpackobj = $packobj->RegData;
	my @filelist = keys(%{$packobj->FileList}); 
	$#filelist != -1 and unpackTGZ($packobj->Untgz,${%{$packobj->MainPath}}{'value'},0,@filelist);
	checkUnpackedArchive(${%{$packobj->MainPath}}{'value'},$packobj->FileList);
	$regpackobj->FileList($packobj->FileList);
	$regpackobj->Script($packobj->Script);
	changeOwner($packobj->User,$packobj->Group,${%{$packobj->MainPath}}{'value'},keys(%{$packobj->FileList}));
}	


sub delete{
	print2stderr("delete function not implemented yet\n");
	diesoft($SAPDB::Install::Values::diemsg);
}					


sub update{
	my ($packobj)=@_;
	my @subnames = ();
	my @subobjs = ();
	$SAPDB::Install::StartInstall::cleaner->SetRef(\@subobjs);
	foreach my $subpackobj (@{$packobj->SubPackObjs}){
		${$$subpackobj->Skip} and next;
		$$subpackobj->Update or next;
		push @subnames,$$subpackobj->Name;
		$$subpackobj->RegData->evalScript();
		if(defined $$subpackobj->RegData->unregister){
			$$subpackobj->RegData->setValid(0);
			$$subpackobj->RegData->Registry->Log->SetMsg("unregister subpackage ".$$subpackobj->DispName."\n");
			eval{
				&{$$subpackobj->RegData->unregister};
			}; $@ and print2stderr("WRN: error in unregister() of subpackage \"".$$subpackobj->DispName."\": $@\n");
		}
	}
	my @all_other_subs;
	foreach $subpack_id (@{$packobj->RegData->SubPackages}){
		my $found = 0;
		foreach my $name (@subnames){
			$subpack_id->[0] eq $name and $found=1 and last;
		}
		unless($found){
			push @all_other_subs,$subpack_id;
		}
	}
	

	my $regpackobj=$packobj->RegData;
	foreach my $subpack_id (@all_other_subs){
		my $subregpackobj = $regpackobj->Registry->getPackage(@$subpack_id,0,1);
		defined $subregpackobj or next;
		$subregpackobj->evalScript();
		if(defined $subregpackobj->unregister){
			$subregpackobj->setValid(0);
			$subregpackobj->Registry->Log->SetMsg("unregister subpackage ".$subregpackobj->DispName."\n");
			eval{
				&{$subregpackobj->unregister};
			}; $@ and print2stderr("WRN: error in unregister() of subpackage \"".$subregpackobj->DispName."\": $@\n");
		}
		push @subobjs,\$subregpackobj;	
		
	}
	
	
	$regpackobj->evalScript();
	if(defined $regpackobj->unregister && (!$packobj->IsSubPackage || $packobj->IsSubPackage && !$packobj->WithParent)){
		eval{
			&{$regpackobj->unregister};
		}; $@ and print2stderr("WRN: error in unregister(): $@\n");
		if($^O=~/aix/i){
			#on aix shared libs may stay in memory after ending related processes
			#slibclean remove all unused shared objects 
			sub ignore_all_errors{
				my ($text)=@_;
				$SAPDB::Install::StartInstall::log->SetMsg("MSG: SYS: slibclean: $text\n");
				return 1;
			}
			callsystem('slibclean',\&ignore_all_errors,1);
		}
	}
	my %newmd5sums = %{$packobj->FileList}; # md5sums of files in tgz archive
	my @newlist=keys(%newmd5sums);
	my %regmd5sums = %{$regpackobj->FileList};
	if(%regmd5sums){# md5sums from registered filelist; original status
		my %curmd5sums=getcurmd5sums($regpackobj->Path,keys(%regmd5sums)); # md5sums of current installed files
		my @reglist=keys(%regmd5sums);
		my @nonew=minus(\@reglist,\@newlist);
		my @modified;
		my @missed;
		foreach my $file (@reglist){
			unless(-f $regpackobj->Path.'/'.$file){ 
				TraceMsg("file \"$file\" not found\n",3,\$DEBUG);
				push @missed,$file;
				next;
			}
			if($regmd5sums{$file} ne $curmd5sums{$file}){
				TraceMsg("file \"$file\" modified\n",3,\$DEBUG);
				push(@modified,$file); # all files of current installation, which was modified
			}
		}
		my @delete = minus(\@nonew,\@modified); # files to delete
		@delete = minus(\@delete,\@missed); # a missed file can't be deleted
		deleteFiles($regpackobj->Path,@delete);
		my @rename = intersect(\@newlist,\@modified); # files to rename
		my @renamed = renameModFiles($regpackobj->Path,@rename);
	}
	$#newlist != -1 and unpackTGZ($packobj->Untgz,$regpackobj->Path,0,@newlist);
	checkUnpackedArchive($regpackobj->Path,$packobj->FileList);
	$regpackobj->FileList(\%newmd5sums);
	$regpackobj->Script($packobj->Script);
	changeOwner($packobj->User,$packobj->Group,${%{$packobj->MainPath}}{'value'},keys(%{$packobj->FileList}));
	return $#subobjs > -1 ? \@subobjs : undef; 
}
	
				
sub patch{
	my ($packobj)=@_;
	my $regpackobj=$packobj->RegData;
	my %newmd5sums = %{$packobj->FileList}; # md5sums of files in tgz archive
	my %regmd5sums = %{$regpackobj->FileList};
	my %curmd5sums=getcurmd5sums(${%{$packobj->MainPath}}{'value'},keys(%newmd5sums)); # md5sums of current installed files
	my @modified;
	foreach my $file (keys(%newmd5sums)){
		if($regmd5sums{$file} ne $curmd5sums{$file}){
			TraceMsg("file \"$file\" modified\n",4,\$DEBUG);
			push(@modified,$file); # all files of current installation, which was modified
		}
	}
	my @rename = intersect(\@newlist,\@modified); # files to rename
	my @renamed = renameModFiles(${%{$packobj->MainPath}}{'value'},@rename);
	$#newlist != -1 and unpackTGZ($packobj->Untgz,${%{$packobj->MainPath}}{'value'},0,@newlist);
	checkUnpackedArchive(${%{$packobj->MainPath}}{'value'},$packobj->FileList);
	$regpackobj->mergeList(%newmd5sums);
	changeOwner($packobj->User,$packobj->Group,${%{$packobj->MainPath}}{'value'},keys(%{$packobj->FileList}));
}
		
sub verify{
	my ($regpackobj)=@_;
	my %regmd5sums=%{$regpackobj->FileList};
	%regmd5sums or print2stderr("cannot verify files in directory \"$path\" - no files registered\n") and return 0;
	my %curmd5sums=getcurmd5sums($regpackobj->Path,keys(%regmd5sums));
	%curmd5sums or print2stderr("cannot verify files in directory \"$path\" - no files found\n") and return 0;
	my $missing=0;
	my $modified=0;
	foreach my $file (keys(%regmd5sums)){
		$curmd5sums{$file} or print2stdout("file \"$file\" missing\n") and $missing++ and next;
		unless($curmd5sums{$file} eq $regmd5sums{$file}){
			print2stdout("file \"$file\" modified\n");
			$modified++;
		}
	}
	print2stdout("verify statistics:\n");
	print2stdout("\t\t$missing file(s) missing\n");
	print2stdout("\t\t$modified file(s) modified\n");
	return 1;
}	
		

sub listUpdatables{
			my ($hrRegisteredPackages,$packobj,$batch)=@_;
			my $text;
			$packobj->Type eq 'PATCH' and $text = 'patch' or $text='update';
			my %registeredPackages = %$hrRegisteredPackages;
			my @show_patchable;
			my @patchable_pathes;
			foreach my $path (keys(%registeredPackages)){
					my $version=$registeredPackages{$path};
					#print "TEST: $version - ".$packobj->Version.." - ".$packobj->MinVersion."\n";
					if(release2num($version) <= release2num($packobj->Version) and release2num($version) >= release2num($packobj->MinVersion)){
						push(@patchable_pathes,$path);
						push(@show_patchable,[$path,$registeredPackages{$path}]);

					}
					else{
						push (@nonpatchable,$path);
					}
			}
			
			$#nonpatchable > -1 and print2stdout("cannot $text following installation(s):\n");
			foreach my $path (@nonpatchable){
				printTable([[$path,$registeredPackages{$path}]],' ');
			}
			print2stdout("\n\n");
			$text =~ s/e$//g; #remove e from update
			$#patchable_pathes == -1 and print2stdout("no ${text}able installation of package \"".$packobj->DispName."\" found\n") and return 0;
			
			#print2stdout("\n\n\nfound  ${text}able installation(s): \n");
			#my $count=0;
			#foreach my $path (@patchable_pathes){
			#	print2stdout($count++.": $path - $registeredPackages{$path}\n");
			#}
			#my $selection;
			#if($count == 0){
			#	return;
			#}
			#elsif($batch){
			#	print2stderr("cannot select ${text}able installation - run in batch mode\n");
			#	diesoft($SAPDB::Install::Values::diemsg);
			#}
			#else{
			#	$selection=ask4update($count);
			#}
			
			$selection=ask4any(\@show_patchable,$text.'able installation','select',$batch);
			$selection == -1 and return;
			return $patchable_pathes[$selection];
}

sub selectInstallation{
	my ($instRegistry_ref,$packobj,$batch)=@_;
	my $instRegistry = $$instRegistry_ref;
    my %registeredPackages;
    if(defined $instRegistry){
		%registeredPackages=$instRegistry->getInstallPathes($packobj->Name);
    }
	my $selectedPath=listUpdatables(\%registeredPackages,$packobj,$batch);
	unless($selectedPath){
		$packobj->Update(0);
		return 1;
	}
	$packobj->Update(1);
	local *mainpath=$packobj->MainPath;
	$mainpath{'value'}=$selectedPath;
	return 1;
}

1;